(* $Id: mlbrot.ml,v 1.156 2008/12/22 09:52:28 deraugla Exp $ *)

open Rt;
open Printf;

type num_type = [ N_int | N_i64 | N_flo | N_big ];

module type Mfloat =
  sig
    type t = 'abstract;
    value set_num : num_type -> t -> bool;
    value string_of_num_type : unit -> string;
    value neg : t -> t;
    value twice : t -> t;
    value half : t -> t;
    value mult : t -> t -> t;
    value of_int : int -> t;
    value to_string : t -> string;
    value to_float : t -> float;
    value of_float : float -> t;
    value coord_of_pos : t -> t -> int -> int -> t;
    value is_mandel_point : t -> t -> int -> option int;
  end
;

module Mfloat_int : Mfloat =
  struct
    type t = int;
    value set_num _ _ = True;
    value string_of_num_type () = "int";
    value m = 32768;
    value fm = float m;
    value mult x y =
      if x = 0 || y = 0 then 0
      else
        let x1 = x / m in
        let x2 = x mod m in
        let y1 = y / m in
        let y2 = y mod m in
        x1 * y1 * m + x1 * y2 + x2 * y1 + (x2 * y2) / m
    ;
    value neg x = - x;
    value twice x = 2 * x;
    value half x = x / 2;
    value of_int i = i * m;
    value to_string x = sprintf "%.8g" (float x /. float m);
    value to_float x = float x /. fm;
    value of_float f = truncate (f *. fm +. 0.5);
    value one_half = m / 2;
    value four = 4 * m;
    value coord_of_pos reduc xc x w = xc + reduc * (x - w / 2);
    value is_mandel_point a b nb_it =
      loop 0 0 0 0 0
      where rec loop it xn yn sqr_xn sqr_yn =
        if it >= nb_it then None
        else
          let u = mult xn yn in
          let yn = u + u + b in
          let xn = sqr_xn - sqr_yn + a in
          let sqr_yn = mult yn yn in
          let sqr_xn = mult xn xn in
          if sqr_xn + sqr_yn >= four then Some it
          else loop (it + 1) xn yn sqr_xn sqr_yn
    ;
  end
;

module Mfloat_i64 : Mfloat =
  struct
    type t = int64;
    value set_num _ _ = True;
    value string_of_num_type () = "i64";
    value mf = ldexp 1. 30;
    value m64 = Int64.of_float mf;
    value mult x y =
      if x = Int64.zero || y = Int64.zero then Int64.zero
      else
        let x1 = Int64.div x m64 in
        let x2 = Int64.rem x m64 in
        let y1 = Int64.div y m64 in
        let y2 = Int64.rem y m64 in
        Int64.add
          (Int64.add
             (Int64.add
                (Int64.mul
                   (Int64.mul x1 y1)
                   m64)
                (Int64.mul x1 y2))
             (Int64.mul x2 y1))
          (Int64.div (Int64.mul x2 y2) m64)
    ;
    value int64_two = Int64.of_int 2;
    value neg = Int64.neg;
    value twice x = Int64.mul x int64_two;
    value half x = Int64.div x int64_two;
    value of_int x = Int64.mul (Int64.of_int x) m64;
    value to_string x = sprintf "%.16g" (Int64.to_float x /. mf);
    value to_float x = Int64.to_float x /. mf;
    value of_float f = Int64.of_float (f *. mf);
    value one_half = Int64.div m64 (Int64.of_int 2);
    value four = Int64.mul (Int64.of_int 4) m64;
    value coord_of_pos reduc xc x w =
      Int64.add xc
        (Int64.sub
           (Int64.mul (Int64.of_int x) reduc)
           (Int64.div (Int64.mul (Int64.of_int w) reduc) int64_two))
    ;
    value is_mandel_point a b nb_it =
      loop 0 Int64.zero Int64.zero Int64.zero Int64.zero
      where rec loop it xn yn sqr_xn sqr_yn =
        if it >= nb_it then None
        else
          let u = mult xn yn in
          let yn = Int64.add (Int64.add u u) b in
          let xn = Int64.add (Int64.sub sqr_xn sqr_yn) a in
          let sqr_yn = mult yn yn in
          let sqr_xn = mult xn xn in
          if Int64.add sqr_xn sqr_yn >= four then Some it
          else loop (it + 1) xn yn sqr_xn sqr_yn
    ;
  end
;

module Mfloat_flo : Mfloat =
  struct
    type t = float;
    value set_num _ _ = True;
    value string_of_num_type () = "flo";
    value neg x = -. x;
    value twice x = 2. *. x;
    value half x = x /. 2.;
    value mult x y = x *. y;
    value of_int = float;
    value to_string x = sprintf "%.16g" x;
    value to_float x = x;
    value of_float x = x;
    value coord_of_pos reduc xc x w =
      xc +. reduc *. (float x -. float w *. 0.5)
    ;
    value is_mandel_point a b nb_it =
      loop 0 0. 0. 0. 0.
      where rec loop it xn yn sqr_xn sqr_yn =
        if it >= nb_it then None
        else
          let u = xn *. yn in
          let xn = sqr_xn -. sqr_yn +. a in
          let yn = u +. u +. b in
          let sqr_xn = xn *. xn in
          let sqr_yn = yn *. yn in
          if sqr_xn +. sqr_yn >= 4. then Some it
          else loop (it + 1) xn yn sqr_xn sqr_yn
    ;
  end
;

module Mfloat_big : Mfloat =
  struct
    open Big_int;
    type t = big_int;
    value set_num _ _ = True;
    value string_of_num_type () = "big";
    type scale =
      { e : mutable int;
        m : mutable big_int;
        half_m : mutable big_int;
        one : mutable big_int;
        b_four : mutable big_int;
        one_half : mutable big_int }
    ;
    value two_big_int = big_int_of_int 2;
    value s =
      {e = 0; m = zero_big_int; half_m = zero_big_int; one = zero_big_int;
       b_four = zero_big_int; one_half = zero_big_int}
    ;
    value set_scale e = do {
      s.e := e;
      s.m := power_int_positive_int 2 e;
      s.half_m := power_int_positive_int 2 (e - 1);
      s.one := s.m;
      s.b_four := mult_int_big_int 4 (square_big_int s.m);
      s.one_half := div_big_int s.one two_big_int;
    };
    set_scale 127;

    value neg = minus_big_int;
    value twice x = mult_int_big_int 2 x;
    value half x = div_big_int x two_big_int;
    value mult x y = div_big_int (mult_big_int x y) s.m;
    value of_int i = mult_int_big_int i s.m;
    value to_string x = sprintf "%.16g" (ldexp (float_of_big_int x) (-s.e));
    value to_float x = ldexp (float_of_big_int x) (-s.e);
    value of_float f =
      let is_neg = f < 0.0 in
      let f = ldexp (abs_float f) s.e in
      loop f [] where rec loop f cl =
        if f = 0.0 then
          loop (big_int_of_int 0) cl where rec loop x =
            fun
            [ [c :: cl] -> loop (add_int_big_int c (mult_int_big_int 10 x)) cl
            | [] -> if is_neg then minus_big_int x else x ]
        else
          let (frac, int) = modf (f /. 10.0) in
          loop int [truncate (10.0 *. frac +. 0.5) :: cl]
    ;
    value coord_of_pos reduc xc x w =
      add_big_int xc
        (sub_big_int
           (mult_int_big_int x reduc)
           (div_big_int (mult_int_big_int w reduc) two_big_int))
    ;
    value is_mandel_point a b nb_it =
      loop 0 zero_big_int zero_big_int zero_big_int zero_big_int
      where rec loop it xn yn b_sqr_xn b_sqr_yn =
        if it >= nb_it then None
        else
          let yn =
            add_big_int (div_big_int (mult_big_int xn yn) s.half_m) b
          in
          let xn =
            add_big_int (div_big_int (sub_big_int b_sqr_xn b_sqr_yn) s.m) a
          in
          let b_sqr_yn = square_big_int yn in
          let b_sqr_xn = square_big_int xn in
          if ge_big_int (add_big_int b_sqr_xn b_sqr_yn) s.b_four then
            Some it
          else
            loop (it + 1) xn yn b_sqr_xn b_sqr_yn
    ;
  end
;

module Mfloat3 : Mfloat =
  struct
    type comb =
      [ F_int of Mfloat_int.t
      | F_i64 of Mfloat_i64.t
      | F_flo of Mfloat_flo.t
      | F_big of Mfloat_big.t ]
    ;
    type t = { a : mutable comb };
    value gnt = ref N_flo;
    value string_of_num_type () =
      match gnt.val with
      [ N_int -> Mfloat_int.string_of_num_type ()
      | N_i64 -> Mfloat_i64.string_of_num_type ()
      | N_flo -> Mfloat_flo.string_of_num_type ()
      | N_big -> Mfloat_big.string_of_num_type () ]
    ;
    value int_value x =
      match x.a with
      [ F_int a -> a
      | F_i64 a -> Mfloat_int.of_float (Mfloat_i64.to_float a)
      | F_flo a -> Mfloat_int.of_float (Mfloat_flo.to_float a)
      | F_big a -> Mfloat_int.of_float (Mfloat_big.to_float a) ]
    ;
    value i64_value x =
      match x.a with
      [ F_int a -> Mfloat_i64.of_float (Mfloat_int.to_float a)
      | F_i64 a -> a
      | F_flo a -> Mfloat_i64.of_float (Mfloat_flo.to_float a)
      | F_big a -> Mfloat_i64.of_float (Mfloat_big.to_float a) ]
    ;
    value flo_value x =
      match x.a with
      [ F_int a -> Mfloat_flo.of_float (Mfloat_int.to_float a)
      | F_i64 a -> Mfloat_flo.of_float (Mfloat_i64.to_float a)
      | F_flo a -> a
      | F_big a -> Mfloat_flo.of_float (Mfloat_big.to_float a) ]
    ;
    value big_value x =
      match x.a with
      [ F_int a -> Mfloat_big.of_float (Mfloat_int.to_float a)
      | F_i64 a -> Mfloat_big.of_float (Mfloat_i64.to_float a)
      | F_flo a -> Mfloat_big.of_float (Mfloat_flo.to_float a)
      | F_big a -> a ]
    ;
    value to_int x =
      match x.a with
      [ F_int a -> a
      | _ -> do {
          let a = int_value x in
          x.a := F_int a;
          a
        } ]
    ;
    value to_i64 x =
      match x.a with
      [ F_i64 a -> a
      | _ -> do {
          let a = i64_value x in
          x.a := F_i64 a;
          a
        } ]
    ;
    value to_flo x =
      match x.a with
      [ F_flo a -> a
      | _ -> do {
          let a = flo_value x in
          x.a := F_flo a;
          a
        } ]
    ;
    value to_big x =
      match x.a with
      [ F_big a -> a
      | _ -> do {
          let a = big_value x in
          x.a := F_big a;
          a
        } ]
    ;
    value neg x =
      match x.a with
      [ F_int a -> {a = F_int (Mfloat_int.neg a)}
      | F_i64 a -> {a = F_i64 (Mfloat_i64.neg a)}
      | F_flo a -> {a = F_flo (Mfloat_flo.neg a)}
      | F_big a -> {a = F_big (Mfloat_big.neg a)} ]
    ;
    value twice x =
      match gnt.val with
      [ N_int -> {a = F_int (Mfloat_int.twice (to_int x))}
      | N_i64 -> {a = F_i64 (Mfloat_i64.twice (to_i64 x))}
      | N_flo -> {a = F_flo (Mfloat_flo.twice (to_flo x))}
      | N_big -> {a = F_big (Mfloat_big.twice (to_big x))} ]
    ;
    value half x =
      match gnt.val with
      [ N_int -> {a = F_int (Mfloat_int.half (to_int x))}
      | N_i64 -> {a = F_i64 (Mfloat_i64.half (to_i64 x))}
      | N_flo -> {a = F_flo (Mfloat_flo.half (to_flo x))}
      | N_big -> {a = F_big (Mfloat_big.half (to_big x))} ]
    ;
    value mult x y =
      match gnt.val with
      [ N_int -> {a = F_int (Mfloat_int.mult (to_int x) (to_int y))}
      | N_i64 -> {a = F_i64 (Mfloat_i64.mult (to_i64 x) (to_i64 y))}
      | N_flo -> {a = F_flo (Mfloat_flo.mult (to_flo x) (to_flo y))}
      | N_big -> {a = F_big (Mfloat_big.mult (to_big x) (to_big y))} ]
    ;
    value of_int i =
      match gnt.val with
      [ N_int -> {a = F_int (Mfloat_int.of_int i)}
      | N_i64 -> {a = F_i64 (Mfloat_i64.of_int i)}
      | N_flo -> {a = F_flo (Mfloat_flo.of_int i)}
      | N_big -> {a = F_big (Mfloat_big.of_int i)} ]
    ;
    value to_string x =
      match gnt.val with
      [ N_int -> Mfloat_int.to_string (to_int x)
      | N_i64 -> Mfloat_i64.to_string (to_i64 x)
      | N_flo -> Mfloat_flo.to_string (to_flo x)
      | N_big -> Mfloat_big.to_string (to_big x) ]
    ;
    value to_float x =
      match x.a with
      [ F_int a -> Mfloat_int.to_float a
      | F_i64 a -> Mfloat_i64.to_float a
      | F_flo a -> Mfloat_flo.to_float a
      | F_big a -> Mfloat_big.to_float a ]
    ;
    value of_float f =
      match gnt.val with
      [ N_int -> {a = F_int (Mfloat_int.of_float f)}
      | N_i64 -> {a = F_i64 (Mfloat_i64.of_float f)}
      | N_flo -> {a = F_flo (Mfloat_flo.of_float f)}
      | N_big -> {a = F_big (Mfloat_big.of_float f)} ]
    ;
    value set_num nt check_not_null =
      let v =
        match nt with
        [ N_int -> Mfloat_int.to_float (int_value check_not_null)
        | N_i64 -> Mfloat_i64.to_float (i64_value check_not_null)
        | N_flo -> Mfloat_flo.to_float (flo_value check_not_null)
        | N_big -> Mfloat_big.to_float (big_value check_not_null) ]
      in
      if v = 0. then False else do { gnt.val := nt; True }
    ; 
    value coord_of_pos reduc xc x w =
      match gnt.val with
      [ N_int ->
          let reduc = to_int reduc in
          {a = F_int (Mfloat_int.coord_of_pos reduc (to_int xc) x w)}
      | N_i64 ->
          let reduc = to_i64 reduc in
          {a = F_i64 (Mfloat_i64.coord_of_pos reduc (to_i64 xc) x w)}
      | N_flo ->
          let reduc = to_flo reduc in
          {a = F_flo (Mfloat_flo.coord_of_pos reduc (to_flo xc) x w)}
      | N_big ->
          let reduc = to_big reduc in
          {a = F_big (Mfloat_big.coord_of_pos reduc (to_big xc) x w)} ]
    ;
    value is_mandel_point a b nb_it =
      match gnt.val with
      [ N_int -> Mfloat_int.is_mandel_point (to_int a) (to_int b) nb_it
      | N_i64 -> Mfloat_i64.is_mandel_point (to_i64 a) (to_i64 b) nb_it
      | N_flo -> Mfloat_flo.is_mandel_point (to_flo a) (to_flo b) nb_it
      | N_big -> Mfloat_big.is_mandel_point (to_big a) (to_big b) nb_it ]
    ;
  end
;

module Mfloat = Mfloat3;

type nedeque 'a =
  { bef : list 'a;
    cur : 'a;
    aft : list 'a }
;

type info =
  { reduc : mutable Mfloat.t;
    xc : mutable Mfloat.t;
    yc : mutable Mfloat.t;
    nb_it : mutable int;
    exposing : mutable bool;
    exposing_y_in_last_chunk : mutable int;
    dont_expose_more : mutable bool;
    no_main_loop : mutable bool;
    level : mutable int;
    random_for_colors : Random.State.t;
    random : mutable nedeque Random.State.t;
    counts_array : array (array int);
    pixmap : pixmap;
    c_red : color;
    c_green : color;
    c_black : color;
    c_white : color;
    c_col : mutable palette;
    watch : mouse }
and palette =
  { t_col : array color_hsv;
    ini_max_s : int;
    ini_max_v : int;
    max_ds : mutable int;
    max_dv : mutable int }
and color_hsv =
  { color : color;
    hsv : (int * int * int) }
;

value mfloat_zero = Mfloat.of_int 0;

value (info, get_info) = user_info "info" (ref UNone);

value user_intervention xa info =
  if rt_pending_events xa then do {
    info.dont_expose_more := False;
    while rt_pending_events xa && not info.dont_expose_more do {
      rt_treat_one_event xa
    };
    info.dont_expose_more
  }
  else False
;

value expose_mandel xa wid x y w h update_nb_it ini_chunk = do {
  let info = get_info wid in
  let drw = PixmapDr info.pixmap in
  let wid_w = widget_width wid in
  let wid_h = widget_height wid in
  let last_col = ref info.c_red in
  rt_select_mouse wid info.watch;
  let exposing = info.exposing in
  let exposing_y_in_last_chunk = info.exposing_y_in_last_chunk in
  info.exposing := True;
  info.exposing_y_in_last_chunk := 0;
  let counts_array = info.counts_array in
  let all_neighbours_same parent_i parent_j d =
    let v = counts_array.(parent_i).(parent_j) in
    parent_i >= d && parent_j >= d &&
    parent_i + d < w && parent_j + d < h &&
    counts_array.(parent_i-d).(parent_j-d) = v &&
    counts_array.(parent_i-d).(parent_j) = v &&
    counts_array.(parent_i-d).(parent_j+d) = v &&
    counts_array.(parent_i).(parent_j-d) = v &&
    counts_array.(parent_i).(parent_j+d) = v &&
    counts_array.(parent_i+d).(parent_j-d) = v &&
    counts_array.(parent_i+d).(parent_j) = v &&
    counts_array.(parent_i+d).(parent_j+d) = v
  in
  let nb_it = info.nb_it in
  let n_bef_it = ref 0 in
  let dist_to_nb_it = ref 0 in
  let ini_chunk =
    match ini_chunk with
    [ Some c -> c
    | None -> 16 ]
  in
  let start_time = ref (Some (info.level, Unix.gettimeofday ())) in
  loop_chunk ini_chunk nb_it
  where rec loop_chunk chunk_size nb_it =
    loop_j 0 where rec loop_j j =
      if j >= h then
        if chunk_size = 1 then ()
        else
          let (chunk_size, nb_it) =
            if chunk_size = 16 && update_nb_it then do {
              let av =
                if n_bef_it.val = 0 then 0
                else
                  truncate
                    (float dist_to_nb_it.val /. float n_bef_it.val +. 0.5)
              in
              if av < 100 then do {
                if av = 0 then info.nb_it := nb_it * 2 else ();
                n_bef_it.val := 0;
                dist_to_nb_it.val := 0;
                (chunk_size, nb_it * 2)
              }
              else do {
                if av > 1500 then info.nb_it := nb_it / 2 else ();
                (chunk_size / 2, nb_it)
              }
            }
            else (chunk_size / 2, nb_it)
          in
          loop_chunk chunk_size nb_it
      else do {
        if chunk_size = 1 then info.exposing_y_in_last_chunk := j else ();
        let parent_j =
          if j mod (2 * chunk_size) = 0 then j
          else j - chunk_size
        in
        loop_i 0 where rec loop_i i =
          if i >= w then do {
            rt_clear_area wid (0, y + j - chunk_size/2, wid_w, chunk_size);
            if user_intervention xa info then start_time.val := None
            else loop_j (j + chunk_size)
          }
          else do {
            let parent_i =
              if i mod (2 * chunk_size) = 0 then i
              else i - chunk_size
            in
            let do_evaluate =
              if chunk_size = 16 then True
              else if i = parent_i && j = parent_j then False
              else if
                all_neighbours_same parent_i parent_j (chunk_size * 2)
              then False
              else True
            in
            let count =
              if do_evaluate then do {
                let x = x + i in
                let y = y + j in
                let a = Mfloat.coord_of_pos info.reduc info.xc x wid_w in
                let b = Mfloat.coord_of_pos info.reduc info.yc y wid_h in
                let count =
                  match Mfloat.is_mandel_point a b nb_it with
                  [ Some n -> do {
                      incr n_bef_it;
                      dist_to_nb_it.val := dist_to_nb_it.val + nb_it - n;
                      n + 2
                    }
                  | None ->
                      1 ]
                in
                counts_array.(i).(j) := count;
                let c =
                  if count > 1 then
                    let n = (count - 2) mod Array.length info.c_col.t_col in
                    info.c_col.t_col.(n).color
                  else
                    info.c_black
                in
                if color_pixel c = color_pixel last_col.val then ()
                else do {
                  rt_select_color c;
                  last_col.val := c;
                };
                if chunk_size = 1 then rt_draw_point drw (x, y)
                else
                  rt_fill_rectangle drw
                    (x - chunk_size / 2, y - chunk_size / 2,
                     chunk_size, chunk_size);
                count
              }
              else do {
                let count = counts_array.(parent_i).(parent_j) in
                counts_array.(i).(j) := count;
                count
              }
            in
            (* for possible color or saturation change while exposing *)
            for k = 0 to chunk_size - 1 do {
              for l = 0 to chunk_size - 1 do {
                let i = i - chunk_size / 2 + k in
                let j = j - chunk_size / 2 + l in
                if i >= 0 && j >= 0 then counts_array.(i).(j) := count else ()
              }
            };
            loop_i (i + chunk_size)
          };
      };
  rt_unselect_mouse wid;
  rt_select_color info.c_black;
  info.exposing := exposing;
  info.exposing_y_in_last_chunk := exposing_y_in_last_chunk;
  info.dont_expose_more := True;
  match start_time.val with
  [ Some (lev, tm1) -> do {
      let tm2 = Unix.gettimeofday () in
      printf "level %d time: %g\n" lev (tm2 -. tm1);
      flush stdout;
    }
  | None -> () ]
};

value random_rotate rs cl =
  rotate_cl (Random.State.int rs (List.length cl)) [] cl
  where rec rotate_cl n rev_beg_cl =
    fun
    [ [c :: cl] ->
        if n = 0 then [c :: cl] @ List.rev rev_beg_cl
        else rotate_cl (n - 1) [c :: rev_beg_cl] cl
    | [] -> assert False ]
;

(* step 0: sat = s2 (max), bri = b2 (max), rand col
   step 1: dec sat from s2 to s1 (=> becoming B&W)
   step 2: change col inc sat from s1 to s2 (=> being col again with anot col)
   step 3: dec bri from b2 to b1 (=> darkened)
   step 4: change col inc bri from b1 to b2 (=> being bri again with anot col)
   continue to step 1 *)

value reinit_colors xd rs old_pal = do {
  Array.iter (fun c -> rt_free_color c.color) old_pal.t_col;
  let min_nb_colors = 350 in
  let s1 = 0 in
  let s2 = 128 in
  let v1 = 0 in
  let v2 = 224 in
  let inc_s = 8 in
  let inc_v = 8 in
  let (s, v, step) =
    match Random.State.int rs 2 with
    [ 0 -> (s1, v2, 2)
    | _ -> (s2, v1, 4) ]
  in
  let max_s = max 0 (min 255 (s2 + old_pal.max_ds)) in
  let max_v = max 0 (min 255 (v2 + old_pal.max_dv)) in
  let stop_step = step in
  loop [] min_nb_colors (Random.State.int rs 256) s v 0 step
  where rec loop cl remaining_colors h s v ostep step =
    if remaining_colors <= 0 && step = stop_step && ostep <> step then
      let cl = random_rotate rs cl in
      {t_col = Array.of_list cl; ini_max_s = s2; ini_max_v = v2;
       max_ds = old_pal.max_ds; max_dv = old_pal.max_dv}
    else
      let ostep = step in
      let (h, s, v, step) =
        if step = 1 then
          let s = s - inc_s in
          let (h, s, step) =
            if s <= s1 then (Random.State.int rs 256, s1, 2) else (h, s, step)
          in
          (h, s, v, step)
        else if step = 2 then
          let s = s + inc_s in
          let (s, step) = if s >= s2 then (s2, 3) else (s, step) in
          (h, s, v, step)
        else if step = 3 then
          let v = v - inc_v in
          let (h, v, step) =
            if v <= v1 then (Random.State.int rs 256, v1, 4) else (h, v, step)
          in
          (h, s, v, step)
        else
          let v = v + inc_v in
          let (v, step) = if v >= v2 then (v2, 1) else (v, step) in
          (h, s, v, step)
      in
      let c =
        let s = s * max_s / s2 in
        let v = v * max_v / v2 in
        let rgb = Mcolor.rgb_of_hsv (h, s, v) in
        rt_create_color xd rgb
      in
      loop [{color = c; hsv = (h, s, v)} ::  cl] (remaining_colors - 1)
        h s v ostep step
};

value refresh_pixel drw info last_col i j =
  if i < Array.length info.counts_array &&
     j < Array.length info.counts_array.(i)
  then do {
    let count = info.counts_array.(i).(j) in
    let c =
      if count > 1 then
        let n = (count - 2) mod Array.length info.c_col.t_col in
        info.c_col.t_col.(n).color
      else
        info.c_black
    in
    if color_pixel c = color_pixel last_col.val then ()
    else do {
      rt_select_color c;
      last_col.val := c;
    };
    rt_draw_point drw (i, j);
  }
  else ()
;

value tempo () = let _ = Unix.select [] [] [] 0.0005 in ();

value refresh_pixmap_with_random_effect wid info = do {
  if info.exposing then () else rt_select_mouse wid info.watch;
  let w = widget_width wid in
  let h = widget_height wid in
  let last_col = ref info.c_black in
  rt_select_color last_col.val;
  let drw = PixmapDr info.pixmap in
  if Random.int 3 = 0 then
    (* random filling by chunks *)
    let chunk_size = 32 in
    let n_wid = (w + chunk_size - 1) / chunk_size in
    let n_hei = (h + chunk_size - 1) / chunk_size in
    let filled = Array.init n_wid (fun _ -> Array.create n_hei False) in
    loop (n_wid * n_hei) where rec loop rest =
      if rest = 0 then ()
      else do {
        let n = Random.int rest in
        let (i, j) =
          loop n 0 0 where rec loop n i j =
            if j = n_hei then assert False
            else if i = n_wid then loop n 0 (j + 1)
            else if filled.(i).(j) then loop n (i + 1) j
            else if n = 0 then (i, j)
            else loop (n - 1) (i + 1) j
        in
        filled.(i).(j) := True;
        let i = i * chunk_size in
        let j = j * chunk_size in
        loop 0 0 where rec loop k l =
          if l = chunk_size then ()
          else if k = chunk_size then loop 0 (l + 1)
          else do {
            let i = i + k in
            let j = j + l in
            refresh_pixel drw info last_col i j;
            loop (k + 1) l
          };
        rt_clear_area wid (i, j, chunk_size, chunk_size);
        tempo ();
        loop (rest - 1);
      }
  else if Random.int 2 = 1 then
    (* slanted paths *)
    let ord_ij = Random.int 2 in
    let dir = Random.int 2 in
    if ord_ij = 0 then
      let (min_imj, max_imj, inc_imj) =
        if dir = 0 then (w-1, -h+1, -1) else (-h+1, w-1, 1)
      in
      loop_imj min_imj min_imj 0 where rec loop_imj imj i j =
        if imj = max_imj + inc_imj then rt_clear_widget wid
        else if i = w + h then do {
          if imj mod 7 = 0 then do {
            rt_clear_widget wid;
            tempo ();
          }
          else ();
          loop_imj (imj + inc_imj) (imj + inc_imj) 0;
        }
        else do {
          if i >= 0 && j >= 0 then refresh_pixel drw info last_col i j
          else ();
          loop_imj imj (i + 1) (j + 1)
        }
    else
      let (min_ipj, max_ipj, inc_ipj) =
        if dir = 0 then (0, w+h-1, 1) else (w+h-1, 0, -1)
      in
      loop_ipj min_ipj 0 min_ipj where rec loop_ipj ipj i j =
        if ipj = max_ipj + inc_ipj then rt_clear_widget wid
        else if j = 0 then do {
          if ipj mod 7 = 0 then do {
            rt_clear_widget wid;
            tempo ();
          }
          else ();
          loop_ipj (ipj + inc_ipj) 0 (ipj + inc_ipj);
        }
        else do {
          if i < w && j < h then refresh_pixel drw info last_col i j
          else ();
          loop_ipj ipj (i + 1) (j - 1)
        }
  else do {
    (* horizontal or vertical paths *)
    let ord_ij = Random.int 2 in
    let dir = Random.int 2 in
    let (min_i, max_i, inc_i) =
     if dir = 0 then (0, w-1, 1) else (w-1, 0, -1)
    in
    let (min_j, max_j, inc_j) =
      if dir = 0 then (0, h-1, 1) else (h-1, 0, -1)
    in
    loop 0 min_i min_j where rec loop cnt i j =
      if ord_ij = 0 && j = max_j + inc_j ||
         ord_ij = 1 && i = max_i + inc_i
      then rt_clear_widget wid
      else if ord_ij = 0 && i = max_i + inc_i then do {
        if cnt = 6 then do {
          rt_clear_widget wid;
          tempo ();
        }
        else ();
        loop ((cnt + 1) mod 7) min_i (j + inc_j)
      }
      else if ord_ij = 1 && j = max_j + inc_j then do {
        if cnt = 6 then do {
          rt_clear_widget wid;
          tempo ();
        }
        else ();
        loop ((cnt + 1) mod 7) (i + inc_i) min_j
      }
      else do {
        refresh_pixel drw info last_col i j;
        if ord_ij = 0 then loop cnt (i + inc_i) j else loop cnt i (j + inc_j)
      }
  };
  if info.exposing then () else rt_unselect_mouse wid;
};

value refresh_pixmap_fast wid info show_fast = do {
  if show_fast || info.exposing then () else rt_select_mouse wid info.watch;
  let w = widget_width wid in
  let h = widget_height wid in
  let last_col = ref info.c_black in
  rt_select_color last_col.val;
  let drw = PixmapDr info.pixmap in
  loop 0 0 where rec loop i j =
    if j = h then
      if show_fast then () else rt_clear_widget wid
    else if i = w then do {
      if show_fast then rt_clear_area wid (0, j, w, 1) else ();
      loop 0 (j + 1)
    }
    else do {
      refresh_pixel drw info last_col i j;
      loop (i + 1) j
    };
  if show_fast || info.exposing then () else rt_unselect_mouse wid;
};

value change_saturation xa xd wid info inc =
  let ini_s = info.c_col.ini_max_s in
  let old_max_ds = info.c_col.max_ds in
  let old_max_s = max 0 (min 255 (ini_s + old_max_ds)) in
  let new_max_ds = if inc = 0 then 0 else old_max_ds + inc * 16 in
  let new_max_s = max 0 (min 255 (ini_s + new_max_ds)) in
  if new_max_s <> old_max_s then do {
    printf "saturation %d\n" new_max_s;
    flush stdout;
    info.c_col.max_ds := new_max_ds;
    let t_col = info.c_col.t_col in
    let ini_v = info.c_col.ini_max_v in
    let max_v = max 0 (min 255 (ini_v + info.c_col.max_dv)) in
    for i = 0 to Array.length t_col - 1 do {
      let c = t_col.(i).color in
      let (h, s, v) = t_col.(i).hsv in
      rt_free_color c;
      let new_s = s * new_max_s / ini_s in
      let new_v = v * max_v / ini_v in
      let rgb = Mcolor.rgb_of_hsv (h, new_s, new_v) in
      t_col.(i) := {color = rt_create_color xd rgb; hsv = (h, s, v)};
    };
    refresh_pixmap_fast wid info False;
  }
  else ()
;

value change_brightness xa xd wid info inc =
  let ini_v = info.c_col.ini_max_v in
  let old_max_dv = info.c_col.max_dv in
  let old_max_v = max 0 (min 255 (ini_v + old_max_dv)) in
  let new_max_dv = if inc = 0 then 0 else old_max_dv + inc * 16 in
  let new_max_v = max 0 (min 255 (ini_v + new_max_dv)) in
  if new_max_v <> old_max_v then do {
    printf "brightness %d\n" new_max_v;
    flush stdout;
    info.c_col.max_dv := new_max_dv;
    let t_col = info.c_col.t_col in
    let ini_s = info.c_col.ini_max_s in
    let max_s = max 0 (min 255 (ini_s + info.c_col.max_ds)) in
    for i = 0 to Array.length t_col - 1 do {
      let c = t_col.(i).color in
      let (h, s, v) = t_col.(i).hsv in
      rt_free_color c;
      let new_s = s * max_s / ini_s in
      let new_v = v * new_max_v / ini_v in
      assert (new_v < 256);
      let rgb = Mcolor.rgb_of_hsv (h, new_s, new_v) in
      t_col.(i) := {color = rt_create_color xd rgb; hsv = (h, s, v)};
    };
    refresh_pixmap_fast wid info False;
  }
  else ()
;

value draw_internal_rectangle xd wid info = do {
  let w = widget_width wid in
  let h = widget_height wid in
  let drw = WidgetDr wid in
  let r = 5 in
  rt_set_line_width xd 1;
  rt_select_color info.c_black;
  rt_draw_rectangle drw (w/4, h/4, w/2, h/2);
  rt_draw_arc drw (w/2-r, h/2-r, 2*r, 2*r, 0, 360*64);
  rt_select_color info.c_white;
  rt_draw_rectangle drw (w/4-1, h/4-1, w/2+2, h/2+2);
  rt_draw_arc drw (w/2-r-1, h/2-r-1, 2*r+2, 2*r+2, 0, 360*64);
};

value recenter info x y w h = do {
  info.xc := Mfloat.coord_of_pos info.reduc info.xc x w;
  info.yc := Mfloat.coord_of_pos info.reduc info.yc y h;
};

value center_of_subwidget main_wid wid =
  let x = widget_x main_wid in
  let y = widget_y main_wid in
  let w = widget_width main_wid in
  let h = widget_height main_wid in
  (x + w / 2 - widget_width wid / 2, y + h / 2 - widget_height wid / 2)
;

value glop = 600.0;
value magn_of_reduc reduc = 1. /. (reduc *. glop);
value reduc_of_magn magn = 1. /. (magn *. glop);

value print_command info = do {
  printf "%s -z %s %s -m %g -i %d\n" Sys.argv.(0)
    (Mfloat.to_string info.xc) (Mfloat.to_string (Mfloat.neg info.yc))
    (magn_of_reduc (Mfloat.to_float info.reduc)) info.nb_it;
  flush stdout;
};

value display_help wid = do {
  let txt = "
    Zooming
      mouse button 1 zoom centered at mouse position
      mouse button 2 center at mouse position
      mouse button 3 unzoom
      z zoom
      s zoom a little bit
      x unzoom
      , zoom on random interesting point
    Color palette
      r change to random color palette
      left-arrow previous color palette
      right-arrow next color palette
      [ decrease saturation
      ] increase saturation
      = set to medium saturation
      } decrease brightness
      { increase brightness
      - set to medium brightness
    Iterations
      d decrease number of iterations
      u increase number of iterations
    Information
      . draw half sized rectangle around center
      c display color palette
      w display in white interesting points
      / erase information
    Computations
      1 computations with integers
      2 computations with integers 64 bits
      3 computations with floats (default)
      4 computations with big integers
    Misc
      l redraw
      t trace
      h help
      q quit
      o print command to start current vision

    Type <escape> to remove this window."
  in
  let (txt, nrow, ncol) =
    let tab =
      loop 1 where rec loop i =
        if i < String.length txt && txt.[i] = ' ' then loop (i + 1)
        else i
    in
    let b = Buffer.create 1 in
    loop_i 0 0 0 0 where rec loop_i i nrow ccol ncol =
      if i = String.length txt then
        let s = Buffer.contents b in
        (String.sub s 1 (String.length s - 1), nrow, ncol)
      else do {
        Buffer.add_char b txt.[i];
        let (i, nrow, ccol, ncol) =
          if txt.[i] = '\n' then
            let i = if txt.[i+1] = ' ' then i + tab else i + 1 in
            (i, nrow + 1, 0, max ccol ncol)
          else
            (i + 1, nrow, ccol + 1, ncol)
        in
        loop_i i nrow ccol ncol
      }
  in
  let sub_wid =
    rt_create_transient_widget wid "help mlbrot"
      (Some rt_destroy_widget)
      (term_desc [] (nrow, ncol, 0)
         (fun wid ->
            fun
            [ TermEvKeyPress {item = K_Escape} -> rt_destroy_widget wid
            | _ -> () ]))
  in
  let vt_newline_mode = "\027[20h" in
  let vt_hide_cursor = "\027[?35h" in
  term_send sub_wid (vt_newline_mode ^ vt_hide_cursor);
  term_send sub_wid txt;
  let (pos_x, pos_y) = center_of_subwidget wid sub_wid in
  rt_map_transient_widget sub_wid pos_x pos_y
};

value rec merge_uniq x =
  fun
  [ [y :: l] -> if y = x then [y :: l] else [y :: merge_uniq x l]
  | [] -> [x] ]
;

value is_interesting_point info ndiff i j =
  let ica = info.counts_array in
  if i <= 0 || i >= Array.length ica - 1 ||
     j <= 0 || j >= Array.length ica.(0) - 1
  then False
  else
    let c = ica.(i).(j) in
    if c > 1 then
      loop [c] (-1) (-1) where rec loop cl di dj =
        if di = 2 then False
        else if dj = 2 then loop cl (di+1) (-1)
        else
          let c = ica.(i+di).(j+dj) in
          let new_cl = merge_uniq c cl in
          if List.length new_cl = ndiff then True
          else loop new_cl (di+1) dj
    else False
;

value iter_interesting_points info ndiff f =
  let w = Array.length info.counts_array in
  let h = Array.length info.counts_array.(0) in
  loop_ij 1 1 where rec loop_ij i j =
    if j = h - 1 then ()
    else if i = w - 1 then loop_ij 1 (j + 1)
    else do {
      if is_interesting_point info ndiff i j then f i j else ();
      loop_ij (i + 1) j
    }
;
 
value expose_optim xa wid info x y w h =
  if (not info.exposing || info.exposing_y_in_last_chunk > y + h / 4) &&
     x > w / 4 && x < 3 * w / 4 && y > h / 4 && y < 3 * h / 4
  then do {
    (* optimization: using what is already computed *)
    printf "type %s level %d nb_it %d (faster)\n"
      (Mfloat.string_of_num_type ()) info.level info.nb_it;
    flush stdout;
    let nb_it = info.nb_it in
    let n_bef_it = ref 0 in
    let dist_to_nb_it = ref 0 in
    let c =
      Array.init w
        (fun i ->
           let i = i / 2 + x - w / 4 in
           Array.init h
             (fun j -> do {
                let j = j / 2 + y - h / 4 in
                let n = info.counts_array.(i).(j) in
                if n > 1 then do {
                  incr n_bef_it;
                  dist_to_nb_it.val :=
                    dist_to_nb_it.val + nb_it - n - 2;
                }
                else ();
                n
              }))
    in
    let av =
      if n_bef_it.val = 0 then 0
      else
        truncate
          (float dist_to_nb_it.val /. float n_bef_it.val +. 0.5)
    in
    if av < 200 then info.nb_it := nb_it * 2 else ();
    for i = 0 to w - 1 do {
      Array.blit c.(i) 0 info.counts_array.(i) 0 h
    };
    refresh_pixmap_fast wid info True;
    expose_mandel xa wid 0 0 w h True (Some 1);
  }
  else do {
    (* normal case *)
    printf "type %s level %d nb_it %d\n"  (Mfloat.string_of_num_type ())
      info.level info.nb_it;
    flush stdout;
    expose_mandel xa wid 0 0 w h True None
  }
;

value default_init_reduc = reduc_of_magn 0.2;

value action_mandel xa xd wid ev =
  let info = get_info wid in
  let w = widget_width wid in
  let h = widget_height wid in
  match ev with
  [ RawEvKeyPress kev ->
      match kev.item with
      [ K_Ascii 'c' -> do {
          let len = Array.length info.c_col.t_col in
          Printf.printf "%d colors\n" len;
          flush stdout;
          let drw = WidgetDr wid in
          for i = 0 to 5 do {
            rt_select_color
              (if i mod 2 = 0 then info.c_black else info.c_white);
            rt_draw_line drw (i, 0) (i, 100);
          };
          for i = 0 to len - 1 do {
            rt_select_color info.c_col.t_col.(i).color;
            rt_draw_line drw (i+6, 0) (i+6, 100);
          };
          for i = 0 to 5 do {
            rt_select_color
              (if i mod 2 = 0 then info.c_black else info.c_white);
            rt_draw_line drw (i+6+len, 0) (i+6+len, 100);
          };
        }
      | K_Ascii 'd' -> do {
          info.nb_it := max 1 (info.nb_it / 3 * 2);
          Printf.printf "level %d nb_it %d\n" info.level info.nb_it;
          flush stdout;
          expose_mandel xa wid 0 0 w h False None;
        }
      | K_Ascii 'h' -> display_help wid
      | K_Ascii 'l' -> expose_mandel xa wid 0 0 w h True None
      | K_Ascii 'o' -> print_command info
      | K_Ascii 'q' -> do {
          info.dont_expose_more := True;
          info.no_main_loop := True;
          rt_stop_main_loop xa
        }
      | K_Ascii 'r' -> do {
          let rs = info.random_for_colors in
          info.random :=
            {bef = [info.random.cur :: info.random.bef];
             cur = Random.State.copy rs;
             aft = []};
          info.c_col := reinit_colors xd rs info.c_col;
          refresh_pixmap_with_random_effect wid info;
        }
      | K_Ascii 's' ->
          let new_reduc = Mfloat.mult info.reduc (Mfloat.of_float 0.97) in
          if Mfloat.to_float new_reduc = 0. then ()
          else do {
            info.reduc := new_reduc;
            let new_level =
              truncate
                ((log default_init_reduc -.
                    log (Mfloat.to_float new_reduc)) /.
                      log 2. +.
                   0.5)
            in
            info.level := new_level;
            Printf.printf "type %s level %d nb_it %d\n"
              (Mfloat.string_of_num_type ()) info.level info.nb_it;
            flush stdout;
            expose_mandel xa wid 0 0 w h True None;
          }
      | K_Ascii 't' -> do {
          printf "level %d\n" info.level;
          printf "magn %g\n" (magn_of_reduc (Mfloat.to_float info.reduc));
          printf "zc %s %s" (Mfloat.to_string info.xc)
            (Mfloat.to_string (Mfloat.neg info.yc));
          printf " nb_it %d\n" info.nb_it;
          let c = info.c_col in
          let s = max 0 (min 255 (c.ini_max_s + c.max_ds)) in
          let v = max 0 (min 255 (c.ini_max_v + c.max_dv)) in
          printf "nb col %d sat %d bri %d\n" (Array.length info.c_col.t_col)
            s v;
          if info.exposing then do {
            printf "exposing";
            if info.exposing_y_in_last_chunk > 0 then
              printf " last chunk y = %d/%d" info.exposing_y_in_last_chunk h
            else ();
            printf "\n";
          }
          else ();
          flush stdout;
        }
      | K_Ascii 'u' -> do {
          info.nb_it := (max 2 info.nb_it) * 3 / 2;
          Printf.printf "level %d nb_it %d\n" info.level info.nb_it;
          flush stdout;
          expose_mandel xa wid 0 0 w h False None;
        }
      | K_Ascii 'w' -> do {
          rt_select_color info.c_white;
          let drw = WidgetDr wid in
          iter_interesting_points info 4 (fun i j -> rt_draw_point drw (i, j))
        }
      | K_Ascii 'x' -> do {
          info.reduc := Mfloat.twice info.reduc;
          info.level := info.level - 1;
          Printf.printf "type %s level %d nb_it %d\n"
            (Mfloat.string_of_num_type ()) info.level info.nb_it;
          flush stdout;
          expose_mandel xa wid 0 0 w h True None;
        }
      | K_Ascii 'z' ->
          let new_reduc = Mfloat.half info.reduc in
          if Mfloat.to_float new_reduc = 0. then ()
          else do {
            info.reduc := new_reduc;
            info.level := info.level + 1;
            let x = w / 2 in
            let y = h / 2 in
            expose_optim xa wid info x y w h
          }
      | K_Ascii '1' | K_KP_N 1 ->
          if Mfloat.set_num N_int info.reduc then do {
            Printf.printf "type %s level %d nb_it %d\n"
              (Mfloat.string_of_num_type ()) info.level info.nb_it;
           flush stdout;
          }
          else ()
      | K_Ascii '2' | K_KP_N 2 ->
          if Mfloat.set_num N_i64 info.reduc then do {
            Printf.printf "type %s level %d nb_it %d\n"
              (Mfloat.string_of_num_type ()) info.level info.nb_it;
           flush stdout;
          }
          else ()
      | K_Ascii '3' | K_KP_N 3 ->
          if Mfloat.set_num N_flo info.reduc then do {
            Printf.printf "type %s level %d nb_it %d\n"
              (Mfloat.string_of_num_type ()) info.level info.nb_it;
           flush stdout;
          }
          else ()
      | K_Ascii '4' | K_KP_N 4 ->
          if Mfloat.set_num N_big info.reduc then do {
            Printf.printf "type %s level %d nb_it %d\n"
              (Mfloat.string_of_num_type ()) info.level info.nb_it;
           flush stdout;
          }
          else ()
      | K_Ascii '[' -> change_saturation xa xd wid info (-1)
      | K_Ascii ']' -> change_saturation xa xd wid info 1
      | K_Ascii '=' -> change_saturation xa xd wid info 0
      | K_Ascii '{' -> change_brightness xa xd wid info (-1)
      | K_Ascii '}' -> change_brightness xa xd wid info 1
      | K_Ascii '-' -> change_brightness xa xd wid info 0
      | K_Ascii '/' -> rt_clear_widget wid
      | K_Ascii '.' -> draw_internal_rectangle xd wid info
      | K_Ascii ',' -> do {
          let cnt = ref 0 in
          iter_interesting_points info 4 (fun i j -> incr cnt);
          if cnt.val = 0 then ()
          else do {
            cnt.val := Random.int cnt.val;
            try
              iter_interesting_points info 4
                (fun i j ->
                   if cnt.val = 0 then do {
                     recenter info i j w h;
                     info.reduc := Mfloat.half info.reduc;
                     info.level := info.level + 1;
                     printf "type %s level %d nb_it %d\n"
                       (Mfloat.string_of_num_type ()) info.level info.nb_it;
                     flush stdout;
                     expose_mandel xa wid 0 0 w h True None;
                     raise Exit;
                   }
                   else decr cnt)
            with
            [ Exit -> () ];
          }
        }
      | K_Left ->
          match info.random.bef with
          [ [] -> ()
          | [rs :: rsl] -> do {
              info.random :=
                {bef = rsl; cur = Random.State.copy rs;
                 aft = [info.random.cur :: info.random.aft]};
              info.c_col := reinit_colors xd rs info.c_col;
              refresh_pixmap_fast wid info False;
            } ]
      | K_Right ->
          match info.random.aft with
          [ [] -> ()
          | [rs :: rsl] -> do {
              info.random :=
                {bef = [info.random.cur :: info.random.bef];
                 cur = Random.State.copy rs; aft = rsl};
              info.c_col := reinit_colors xd rs info.c_col;
              refresh_pixmap_fast wid info False;
            } ]
      | K_Escape -> rt_clear_widget wid
      | _ -> () ]
  | RawEvButtonPress x y _ _ k ->
      match k.item with
      [ 1 ->
          let new_reduc = Mfloat.half info.reduc in
          if Mfloat.to_float new_reduc = 0. then ()
          else do {
            recenter info x y w h;
            info.reduc := new_reduc;
            info.level := info.level + 1;
            expose_optim xa wid info x y w h;
          }
      | 2 -> do {
          recenter info x y w h;
          expose_mandel xa wid 0 0 w h True None;
        }
      | 3 -> do {
          info.reduc := Mfloat.twice info.reduc;
          info.level := info.level - 1;
          Printf.printf "type %s level %d nb_it %d\n"
            (Mfloat.string_of_num_type ()) info.level info.nb_it;
          flush stdout;
          expose_mandel xa wid 0 0 w h True None;
        }
      | _ -> () ]
  | RawEvConfigureNotify (ow, oh, _) (w, h, _) ->
      if w <> ow || h <> oh then expose_mandel xa wid 0 0 w h True None
      else ()
  | _ ->
      () ]
;

value init_wid = 600;
value init_hei = 600;
value init_xc = ref 0.0;
value init_yc = ref 0.0;
value init_reduc = ref default_init_reduc;
value init_nb_it = ref (64, False);
value ini_seed = ref None;

value usage = "mlbrot [-z xc yc] [-m reduc] [-i nb_it] [-s seed]";

value speclist () =
  Arg.align
    [("-m",
      Arg.Float
        (fun f -> if f <> 0. then init_reduc.val := reduc_of_magn f else ()),
      sprintf "<m> Set initial magnification (default: %.1f)"
        (magn_of_reduc init_reduc.val));
     ("-z",
      Arg.Tuple [Arg.Set_float init_xc; Arg.Set_float init_yc],
      sprintf "<xc><yc> Set initial position (default %.1f %.1f)"
        init_xc.val init_yc.val);
     ("-i", Arg.Int (fun i -> init_nb_it.val := (i, True)),
      sprintf "<int> Set initial number of iterations (default %d)"
        (fst init_nb_it.val));
     ("-s", Arg.Int (fun i -> ini_seed.val := Some i),
      "<int> Set random seed for color palettes")]
;

value anonfun s = raise (Arg.Bad ("don't know what to do with " ^ s));

value main () = do {
  Arg.parse (speclist ()) anonfun usage;
  let xd = rt_initialize "" in
  let xa = rt_args [xd] in
  let wid =
    rt_create_widget xd "mlbrot" "mlbrot" AutoPosition
      (Some (fun _ -> rt_stop_main_loop xa))
      (raw_desc []
         (init_wid, init_hei, 0,
          [SelKeyPress; SelStructureNotify; SelButtonPress])
         (action_mandel xa xd))
  in
  let scr_w = screen_width xd in
  let scr_h = screen_height xd in
  let wid_w = widget_width wid in
  let wid_h = widget_height wid in
  let pixmap = rt_create_pixmap xd scr_w scr_h in
  let c_red = rt_create_color xd (255, 0, 0) in
  let c_green = rt_create_color xd (0, 255, 0) in
  let c_black = rt_black_color xd in
  let c_white = rt_white_color xd in
  let rs =
    let a =
      match ini_seed.val with
      [ Some s -> [| s |]
      | None -> [| 0; 1 |] ]
    in
    Random.State.make a
  in
  let random_cur = Random.State.copy rs in
  let c_col =
    reinit_colors xd rs
      {t_col = [| |]; ini_max_s = 0; ini_max_v = 0; max_ds = 0; max_dv = 0}
  in
  let rs =
    match ini_seed.val with
    [ Some _ -> rs
    | None -> Random.State.make_self_init () ]
  in
  Random.self_init ();
  let watch = rt_create_font_mouse xd xC_watch in
  let nb_it = fst init_nb_it.val in
  let user_nb_it = snd init_nb_it.val in
  let reduc = Mfloat.of_float init_reduc.val in
  let xc = Mfloat.of_float init_xc.val in
  let yc = Mfloat.of_float (-. init_yc.val) in
  let level =
    truncate ((log default_init_reduc -. log init_reduc.val) /. log 2. +. 0.5)
  in
  let counts_array = Array.init scr_w (fun i -> Array.create scr_h 0) in
  let ui =
    {reduc = reduc; xc = xc; yc = yc; nb_it = nb_it;
     exposing = False; exposing_y_in_last_chunk = 0;
     dont_expose_more = False; no_main_loop = False;
     level = level; random_for_colors = rs;
     random = {bef = []; cur = random_cur; aft = []};
     counts_array = counts_array; pixmap = pixmap;
     c_red = c_red; c_green = c_green;
     c_black = c_black; c_white = c_white;
     c_col = c_col; watch = watch}
  in
  rt_set_user_info wid (info ui);
  rt_change_background wid (PixmapPn pixmap);
  rt_map_widget wid;
  let info_wid =
    rt_create_popup_widget xd
      (title_desc [] "type h for help"
         (fun wid ->
            fun
            [ TitleEvButtonPress _ -> do {
                rt_cancel_timeout xa "info";
                rt_destroy_widget wid
              } ]))
  in
  let (x, y) = center_of_subwidget wid info_wid in
  rt_map_popup_widget info_wid x y 0;
  rt_set_timeout xa "info" (rt_current_time xa + 1000)
    (fun () -> rt_destroy_widget info_wid);
  printf "type %s level %d nb_it %d\n"  (Mfloat.string_of_num_type ())
    ui.level ui.nb_it;
  flush stdout;
  expose_mandel xa wid 0 0 wid_w wid_h (not user_nb_it) None;
  if ui.no_main_loop then () else rt_main_loop xa;
  print_command ui;
};

main ();
